/*
** Copyright (C) ARM Limited, 2006-2014. All rights reserved.
*/

/* This file contains the main() program that sets the vector table location, displays a welcome message,
initializes the MPU, starts the SysTick timer, initializes the Process Stack Pointer, changes Thread mode
to Unprivileged and to use the Process Stack, then runs the main application (sorts) */


#include "scs.h"
#include "timer.h"
#include <stdio.h>
#include <math.h>

#define VectorTableOffsetRegister 0xE000ED08
extern unsigned int Image$$VECTORS$$Base;
extern unsigned int Image$$PROCESS_STACK$$ZI$$Limit;

extern int compare_sorts(void);
float calculate( float a, float b);


/* Enable the FPU if required */
#ifdef __TARGET_FPU_VFP
extern void $Super$$__rt_lib_init(void);

void $Sub$$__rt_lib_init(void)
{
    /* Enable the FPU in both privileged and user modes by setting bits 20-23 to enable CP10 and CP11 coprocessors */
    SCS.CPACR = SCS.CPACR | (0xF << 20);
    $Super$$__rt_lib_init();
}
#endif

// Debug Exception Monitor and Control register
#define DEMCR   (*((unsigned long *) 0xe000edfc))
#define TRCENA  0x01000000  // enable trace

// Stimulus Port registers
#define ITM_STIM0 (*((unsigned long *) 0xe0000000))
#define ITM_STIM1 (*((unsigned long *) 0xe0000004))
#define ITM_STIM2 (*((unsigned long *) 0xe0000008))
#define ITM_STIM3 (*((unsigned long *) 0xe000000c))

// Trace enable registers
#define ITM_TER   (*((unsigned long *) 0xe0000e00))

// Privilege register: registers that can be accessed by unprivileged code
#define ITM_TPR   (*((unsigned long *) 0xe0000e40))

// Trace Control register
#define ITM_TCR   (*((unsigned long *) 0xe0000e80))

// Lock Access register
#define ITM_LAR   (*((unsigned long *) 0xe0000fb0))

// unlock value
#define ITM_LAR_ACCESS  0xc5acce55 

void ITM_init(void)
{
    ITM_LAR = ITM_LAR_ACCESS; // unlock
    ITM_TCR = 0x1;            // global enable for ITM
    ITM_TPR = 0x1;            // first 8 stim registers have unpriv access
    ITM_TER = 0xf;            // enable 4 stim ports
    DEMCR = TRCENA;           // global enable DWT and ITM
}

__attribute__((noreturn)) void main(void)
{

    /* Processor starts-up in Privileged Thread Mode using Main Stack */

    /* Named register variables */
    register unsigned int SP_PROCESS __asm("psp");
    register unsigned int CONTROL __asm("control");

    /* Tell the processor the location of the vector table, obtained from the scatter file */
    *(volatile unsigned int *)(VectorTableOffsetRegister) = (unsigned int) &Image$$VECTORS$$Base;

    /* Initialize MPU */
    SCS_init();

    /* Initialize ITM */
    ITM_init();

    /* Display a welcome message via semihosting */
    printf("Cortex-M4 bare-metal startup example\n");

    /* Perform a float calculation */
#ifdef __TARGET_FPU_VFP
    printf("Calculating using the Cortex-M4's FPU\n");
#else
    printf("Calculating using the software floating point library (no FPU)\n");
#endif
    printf("Float result should be 80.406250\n");
    printf("Float result is        %f\n", calculate(1.0f, 2.5f));

    /* Initialize SysTick Timer */
    SysTick_init();

    /* Run the main application (sorts) */
    compare_sorts();

    ITM_STIM3 = 0xcafec0de;
    while( 1 )
    {
        /* Loop forever */
    }
}


/* Float calculation to demonstrate the Cortex-M4's FPU, including float Fused Multiply Add (fma) */
float calculate( float a, float b)
{
    a = a + 3.25f;
    b = b * 4.75f;
    return fmaf(a + b, a, b);
    // result should be (4.25+11.875)*4.25 + 11.875 = 80.40625
}
